#!/usr/bin/env python3
# coding: utf-8 
#
# STL writer modified by S.Wakashima (2020/09)
# This program is besed on the originall stl_writer.py (MIT License) from 
# https://github.com/telefab/ImageExtruder/blob/master/photomaton/stl_writer.py
# 

import math
import struct
from struct import unpack 
import numpy as np

# {****} will be replaced 
ASCII_FACET = """facet normal {vec[0]:.8e} {vec[1]:.8e} {vec[2]:.8e}
  outer loop
    vertex {face[0][0]:.8e} {face[0][1]:.8e} {face[0][2]:.8e}
    vertex {face[1][0]:.8e} {face[1][1]:.8e} {face[1][2]:.8e}
    vertex {face[2][0]:.8e} {face[2][1]:.8e} {face[2][2]:.8e}
  endloop
endfacet
"""

######### functions for Vectors #################################

#functions for calculation of the normal vector (right-thumb-rule)
def crossProduct(a, b):
    #calculate cross product of two threedimensional vectors
    if (len(a)!=3) or (len(b)!=3):
        raise ValueError('unvalid value')
    return [a[1]*b[2] - a[2]*b[1], a[2]*b[0] - a[0]*b[2], a[0]*b[1] - a[1]*b[0]]

def dotProduct(a, b):
    #calculate dot product of two threedimensional vectors
    if (len(a)!=3) or (len(b)!=3):
        raise ValueError('unvalid value')
    return [a[0]*b[0]+a[1]*b[1]+a[2]*b[2]]

def normalize(a):
    #normalize a threedimensional vector
    if (len(a)!=3):
        raise ValueError('unvalid value')
    length = sqrt(dotProduct(a, a))
    return [a[0]/length, a[1]/length, a[2]/length] 
 
def diff(v1,v2):
    #substracting one list from another
    if (len(v1)!=3) or (len(v2)!=3):
        raise ValueError('unvalid value')
    return [v1[0]-v2[0],v1[1]-v2[1],v1[2]-v2[2]]
 
def normal_vector(v1,v2,v3):
    #calculate normal vector on triangle spanned by the three vertices
    #tells in which direction the plane looks "right-thumb-rule"
    if (len(v1)!=3) or (len(v2)!=3) or (len(v2)!=3):
        raise ValueError('unvalid value')
    n=crossProduct(diff(v2,v1),diff(v3,v1))
    absolut=0
    for i in n:
        if i>=0:
            absolut+=i
        else:
            absolut-=i
    if absolut == 0:
        print("this should not have happened!")
        return n
    else :
        return [n[0]/absolut,n[1]/absolut,n[2]/absolut]

################################################################### 

#simple function to create sets of three vertices
def triangulate(vertices):
    print("triangulate!")
    n=len(vertices)
    if(n==3):
        return vertices
    elif(n>3):
        facets = triangulate(facet)
        self.add_facets(facets)
    else:
        raise ValueError('wrong number of vertices')
 
    def add_facets(self, facets):
        print("add %d facets" % len(facets))
        for facet in facets:
            self.add_facet(facet)
 
def extrude(self,bottom,height):
    if len(bottom) < 3 :
        raise ValueError('not a polygon')
    else :
        top = []
 
        for vertice in bottom :
            top.append([vertice[0],vertice[1],vertice[2]+height])
 
        bottom.reverse()
        self.add_facet(bottom)
        bottom.reverse()
        
        for i in range(0,len(bottom)-1) :
            self.add_facet([bottom[i],bottom[i+1],top[i+1],top[i]])
            self.add_facet([bottom[len(bottom)-1],bottom[0],top[0],top[len(bottom)-1]])
        
        self.add_facet(top)
 
#### Export 3D objects build of 3 or 4 vertices as ASCII/BIN STL file. #####

class ASCII_STL_Writer:
    def __init__(self, stream):
        self.fp = stream
        self._write_header()

    def _write_header(self):
        self.fp.write("solid SOLIDNAME\n")    # "python" allowed to be any string

    def close(self):
        self.fp.write("endsolid SOLIDNAME\n") # "python" should be the same with the header definition

    def _write(self, face):
        V0 = (face[0][0], face[0][1], face[0][2]) 
        V1 = (face[1][0], face[1][1], face[1][2]) 
        V2 = (face[2][0], face[2][1], face[2][2]) 
        vec = normal_vector(V0, V1, V2)
        self.fp.write(ASCII_FACET.format(face=face, vec=vec))

    def _split(self, face):
        p1, p2, p3, p4 = face
        return (p1, p2, p3), (p3, p4, p1)

    def add_face(self, face):
        """ Add one face with 3 or 4 vertices. """
        if len(face) == 4:
            face1, face2 = self._split(face)
            self._write(face1)
            self._write(face2)
        elif len(face) == 3:
            self._write(face)
        else:
            raise ValueError('only 3 or 4 vertices for each face')

    def add_faces(self, faces):
        """ Add many faces. """
        for face in faces:
            self.add_face(face)

class BIN_STL_Writer(ASCII_STL_Writer):
    def __init__(self, stream):
        self.st = stream
        self.count = 0
        self._write_header()
 
    def close(self):
        self._write_header()
 
    def _write_header(self):
        self.st.seek(0)
        #format characters: http://docs.python.org/2/library/struct.html
        self.st.write(struct.pack("80sI", b'Python Binary STL Writer', self.count))

    def _write(self, facet):
        self.count += 1
        print("add facet %d" % self.count)
        n = normal_vector(facet[0],facet[1],facet[2])
        data = [
        	n[0], n[1], n[2],
        	facet[0][0], facet[0][1], facet[0][2],
        	facet[1][0], facet[1][1], facet[1][2],
        	facet[2][0], facet[2][1], facet[2][2],
        	0
                ]
        #format characters: http://docs.python.org/2/library/struct.html
        self.st.write(struct.pack("12fH", *data))

    def add_facet(self, facet):
        if len(facet) == 3:
            self._write(facet)
                #print '[%s]' % ', '.join(map(str, facet))
        elif len(facet) > 3:
            facets = triangulate(facet)
            self.add_facets(facets)
        else:
            raise ValueError('wrong number of vertices')

    def add_facets(self, facets):
        print("add %d facets" % len(facets))
        for facet in facets:
            self.add_facet(facet)

    def extrude(self,bottom,height):
        if len(bottom) < 3 :
            raise ValueError('not a polygon')
        else :
            top = []
            
        for vertice in bottom :
            top.append([vertice[0],vertice[1],vertice[2]+height])

        bottom.reverse()
        self.add_facet(bottom)
        bottom.reverse()

        for i in range(0,len(bottom)-1) :
            self.add_facet([bottom[i],bottom[i+1],top[i+1],top[i]])
            self.add_facet([bottom[len(bottom)-1],bottom[0],top[0],top[len(bottom)-1]])

        self.add_facet(top)

##### Reader 

def ReadAsciiSTL(fname):

   fp = open(fname, 'r')

   # Read header line
   line_header = fp.readline().strip()
   Header = line_header

   # Read facet lines
   facetcount = 0
   record_dtype = np.dtype([
                   ('normals', np.float32,(3,)), 
                   ('Vertex1', np.float32,(3,)),
                   ('Vertex2', np.float32,(3,)),
                   ('Vertex3', np.float32,(3,))
                   ])
   data = np.array([],dtype=record_dtype)

   while True:
       # Read facet normal line
       line_facetnormal = fp.readline()
       if line_facetnormal.startswith('endsolid'):
           break # Finished
       else:
           line = line_facetnormal.strip().split()
           VECN = [float(line[2]), float(line[3]), float(line[4])]
           facetcount +=1
 
           # Read begin loop
           line_begin_loop = fp.readline().strip()
   
           # Read 3 vertices in each FACET and store vertices
           line_vertex = fp.readline() 
           line = line_vertex.strip().split()
           VEC1 = [float(line[1]), float(line[2]), float(line[3])]

           line_vertex = fp.readline() 
           line = line_vertex.strip().split()
           VEC2 = [float(line[1]), float(line[2]), float(line[3])]

           line_vertex = fp.readline() 
           line = line_vertex.strip().split()
           VEC3 = [float(line[1]), float(line[2]), float(line[3])]
           
           # Read two more identifiers
           line_end_loop = fp.readline().strip()
           line_end_facet = fp.readline().strip()
       
           # Check:
           if not line_facetnormal.startswith('facet normal'):
               print('Warning: expected facet normal.')
           if line_begin_loop != "outer loop":
               print('Warning: expected outer loop.')
           if line_end_loop != "endloop":
               print('Warning: expected endloop.')
           if line_end_facet != "endfacet":
               print('Warning: expected endfacet.')

           # Add to vertices list
           newfacetdata = np.array((VECN, VEC1, VEC2, VEC3), dtype=record_dtype)
           data = np.append(data, newfacetdata)

   fp.close()
   print(data)

   Normals = data['normals']
   Vertex1 = data['Vertex1']
   Vertex2 = data['Vertex2']
   Vertex3 = data['Vertex3']
 
   p = np.append(Vertex1,Vertex2,axis=0)
   p = np.append(p,Vertex3,axis=0) #list(v1)
   Points = np.array(list(set(tuple(p1) for p1 in p)))
   
   return Header,Points,Normals,Vertex1,Vertex2,Vertex3


def ReadBinarySTL(fname):
    """
    UINT8[80] – Header
    UINT32 – Number of triangles
    
    foreach triangle
    REAL32[3] – Normal vector
    REAL32[3] – Vertex 1
    REAL32[3] – Vertex 2
    REAL32[3] – Vertex 3
    UINT16 – Attribute byte count
    end
    """
    fp = open(fname, 'rb')
    Header = fp.read(80) # any header string (80byte)
    nn = fp.read(4)      # number of facet (4byte)
    Numtri = unpack('i', nn)[0]
    print(nn)
    record_dtype = np.dtype([
                   ('normals', np.float32,(3,)),
                   ('Vertex1', np.float32,(3,)),
                   ('Vertex2', np.float32,(3,)),
                   ('Vertex3', np.float32,(3,)),
                   ('atttr', '<i2',(1,) )
    ])
    data = np.fromfile(fp , dtype = record_dtype , count = Numtri)
    fp.close()
    print(data)
 
    Normals = data['normals']
    Vertex1 = data['Vertex1']
    Vertex2 = data['Vertex2']
    Vertex3 = data['Vertex3']
 
    p = np.append(Vertex1,Vertex2,axis=0)
    p = np.append(p,Vertex3,axis=0) #list(v1)
    Points =np.array(list(set(tuple(p1) for p1 in p)))
     
    return Header,Points,Normals,Vertex1,Vertex2,Vertex3

#function as main()
def main():

    # Edit and uncomment from here
    
    def get_tetra():
        # vertex points
        s = 1.
        p0 = (0, 0, 0)
        p1 = (s, 0, 0)
        p2 = (0, s, 0)
        p3 = (0, 0, s)
        # define the 4 faces
        return [
            [p0, p2, p1],
            [p0, p3, p2],
            [p0, p1, p3],
            [p1, p2, p3],
        ]

    # Write ASCII STL
    with open('tetra_ascii.stl','w') as fp:
        writer = ASCII_STL_Writer(fp)
        writer.add_faces(get_tetra())
        writer.close()
    
    
    # To here

    """
    def get_cube():
        # cube corner points
        s = 10.
        p0 = (0, 0, 0)
        p1 = (0, 0, s)
        p2 = (0, s, 0)
        p3 = (0, s, s)
        p4 = (s, 0, 0)
        p5 = (s, 0, s)
        p6 = (s, s, 0)
        p7 = (s, s, s)
        # define the 6 cube faces
        # faces just lists of 3 or 4 vertices
        return [
            [p0, p1, p3, p2],
            [p1, p0, p4, p5],
            [p4, p6, p7, p5],
            [p2, p3, p7, p6],
            [p1, p5, p7, p3],
            [p0, p2, p6, p4],
        ]

    # Write ASCII STL
    with open('cube_ascii.stl', 'w') as fp:
        writer = ASCII_STL_Writer(fp)
        writer.add_faces(get_cube())
        writer.close()
    
    # Write Binary STL
    with open('cube_bin.stl', 'wb') as fp:
        writer = BIN_STL_Writer(fp)
        writer.add_faces(get_cube())
        writer.close()

    # Read ASCII STL
    fname = "cube_ascii.stl" 
    head,p,n,v1,v2,v3 = ReadAsciiSTL(fname)
    print(fname,"header: ",head)
    print(p.shape) 
    print(n.shape)
    print(v1.shape)
    print(v2.shape)
    print(v3.shape)
    
    # Read Binary STL
    fname = "cube_bin.stl" 
    head,p,n,v1,v2,v3 = ReadBinarySTL(fname)
    print(fname,"header: ",head)
    print(p.shape) 
    print(n.shape)
    print(v1.shape)
    print(v2.shape)
    print(v3.shape)
    """

if __name__ == '__main__':
    main()
